/*
	FramesToPNGs.c
	Sam Bushell, 23june1999.
*/


// INCLUDES
#include <Traps.h>
#include <Gestalt.h>
#include <FixMath.h>
#include <Fonts.h>

#include <ImageCompression.h>
#include <QuickTimeComponents.h>
#include <MoviesFormat.h>
#include <Movies.h>
#include <ImageCodec.h>
#include <Errors.h>
#include <Events.h>
#include <Files.h>
#include <Script.h>
#include <TextUtils.h>
#include <AppleEvents.h>

// DEFINES
#define	BailNil(n)		if (!n) { if(!result) result = -1; goto bail; }
#define	BailError(n)	if (n) { if(!result) result = -1; goto bail; }
#ifndef IntToFixed
#define IntToFixed(i)	(((Fixed)(i))<<16)
#endif

	/* What's the maximum number of input movies we will collect? */
#define kMaxSpecs	100


//
//	ExportGWorldToImageFile
//
static OSErr ExportGWorldToImageFile( GWorldPtr gw, FSSpec *file )
{
	OSErr err = noErr;
	GraphicsExportComponent ge = 0;
	
	err = OpenADefaultComponent( GraphicsExporterComponentType, kQTFileTypePNG, &ge );
	if( err ) goto bail;
	
	err = GraphicsExportSetInputGWorld( ge, gw );
	if( err ) goto bail;
	
	err = GraphicsExportSetOutputFile( ge, file );
	if( err ) goto bail;
	
	err = GraphicsExportDoExport( ge, nil );
	if( err ) goto bail;
	
bail:
	if( ge ) CloseComponent( ge );
	return err;
}

//
//	Count the number of video "frames" in the movie by stepping through all of the
//	video "interesting times", or in other words, the places where the movie displays
//	a new video sample. The time between these interesting times is not necessarily constant.
//
static OSErr countVideoFramesInMovie( Movie theMovie, long *outFrameCount )
{
	OSType		whichMediaType = VideoMediaType;
	short		flags = nextTimeMediaSample + nextTimeEdgeOK;
	TimeValue	duration;
	TimeValue	theTime = 0;
	long		frameCount;
	
	frameCount = -1;
	while (theTime >= 0) {
		frameCount++;
		GetMovieNextInterestingTime( theMovie, flags, 1, &whichMediaType, 
									 theTime, 0, &theTime, &duration );
		
		//	After the first interesting time, don't include the time we are currently at.
		flags = nextTimeMediaSample;
	}
	
	if( outFrameCount ) 
		*outFrameCount = frameCount;

	return noErr;
}

static void concatenateStringAndNumber( Str255 str, long num, StrFileName outStr )
{
	Str255 numStr;
	
	NumToString( num, numStr );
	BlockMoveData( str, outStr, str[0]+1 );
	BlockMoveData( &numStr[1], &outStr[1+outStr[0]], numStr[0] );
	outStr[0] += numStr[0];
}

static void makeUpFileName( long num, StrFileName outStr )
{
	BlockMoveData( "\pimage 000", outStr, 10 );
	outStr[7] = '0' + ((num / 100) % 10);
	outStr[8] = '0' + ((num / 10) % 10);
	outStr[9] = '0' + (num % 10);
}

//
//	WriteOutEachFrameOfAMovie
//
static void
WriteOutEachFrameOfAMovie( FSSpec *inSpec )
{
	short		theMovieRefNum = 0;
	Movie		theMovie = nil;
	OSErr		result = noErr;
	short		theMovieResID = 0;
	Rect		movieBox;
	GWorldPtr	gw = nil;
	
	result = OpenMovieFile( inSpec, &theMovieRefNum, fsRdWrPerm );
	if( result ) goto bail;
	result = NewMovieFromFile( &theMovie, theMovieRefNum, &theMovieResID, 0, newMovieActive, 0 );
	if( result ) goto bail;

	// Make the movie draw into an offscreen GWorld.
	GetMovieBox( theMovie, &movieBox );
	result = NewGWorld( &gw, 32, &movieBox, nil, nil, 0 );
	
	SetGWorld( gw, nil );
	EraseRect( &gw->portRect );
	SetMovieGWorld( theMovie, gw, nil );

	// Work through each video frame...
	{
		OSType		whichMediaType = VideoMediaType;
		short		flags = nextTimeMediaSample + nextTimeEdgeOK;
		TimeValue	duration;
		TimeValue	theTime = 0;
		long		frameCount;
		FSSpec		imageFile;
		EventRecord event;
		
		frameCount = -1;
		while (theTime >= 0) {
			frameCount++;
			GetMovieNextInterestingTime( theMovie, flags, 1, &whichMediaType, 
										 theTime, 0, &theTime, &duration );
			if( theTime < 0 )
				break;
			
			// Force the movie to render into the offscreen gworld.
			SetMovieTimeValue( theMovie, theTime );
			MoviesTask( theMovie, 0 );
			
			imageFile = *inSpec;
			
			//concatenateStringAndNumber( "\pimage ", frameCount, imageFile.name );
			makeUpFileName( frameCount, imageFile.name );

			result = ExportGWorldToImageFile( gw, &imageFile );
			if( result ) goto bail;

			//	After the first interesting time, don't include the time we are currently at.
			flags = nextTimeMediaSample;
			
			//	Abort if the user clicked the mouse or pressed a key.
			if( EventAvail( keyDownMask | mDownMask, &event ) )
				break;
		}
	}

	
bail:

	if( theMovieRefNum )
		CloseMovieFile( theMovieRefNum );
	if( theMovie )
		DisposeMovie( theMovie );
	if( gw )
		DisposeGWorld( gw );
}

//
//	WriteOutEachFrameOfSomeMovies
//
static void
WriteOutEachFrameOfSomeMovies(
	FSSpec *	inSpecList,
	UInt16		inSpecCount )
{
	UInt16		theSpecIter;
	
	for( theSpecIter = 0; theSpecIter < inSpecCount; theSpecIter++ )
		WriteOutEachFrameOfAMovie( &inSpecList[theSpecIter] );
}

//
//	PromptUserForAFileAndDoIt
//
static void PromptUserForAFileAndDoIt()
{
	OSErr		result;
	FSSpec		theSpec;
	
	Point		where;
	SFTypeList	theTypeList;
	SFReply		theReply;

	where.h = where.v = -2;
	theTypeList[0] = 'MooV';

	SFGetFilePreview( where, "\p", nil, 1, theTypeList, nil, &theReply );
	if (!theReply.good)
		return;

	// Create an FSSpec from the reply information.
	result = FSMakeFSSpec( theReply.vRefNum, 0, theReply.fName, &theSpec );
	BailError(result);
		
	WriteOutEachFrameOfAMovie( &theSpec );
bail:;
}


//====================================================================================

static Boolean gDone = false;

//====================================================================================
//
//	AppleEvent handling.
//

//
//	DoAppleEvent
//
static void 
DoAppleEvent( 
	const EventRecord *inEvent )
{
	OSErr err;

  // should check for your own event message types here - if you have any
	
	err = AEProcessAppleEvent( inEvent );
}

//
//	DoAEOpenApp
//
static pascal OSErr 
DoAEOpenApp(
	const AppleEvent * /*message*/,
	const AppleEvent * /*reply*/,
	long /*refcon*/ )
{
	gDone = true;
	PromptUserForAFileAndDoIt();
	return noErr;
}

//
//	DoAEOpenDocument
//
static pascal OSErr 
DoAEOpenDocument(
	const AppleEvent * message, 
	const AppleEvent * /*reply*/, 
	long /*refcon*/ )
{
	long        index;
	long        itemsInList;
	AEKeyword   keywd;
	OSErr       err;
	OSErr       ignoreErr;
	AEDescList  docList;
	long        actSize;
	DescType    typeCode;
	FSSpec      theFSSpec;
	FSSpec		theSpecList[ kMaxSpecs ];
	UInt16		theSpecCount = 0;

	gDone = true;

	/*open the specified documents*/

	docList.dataHandle = nil;

	err = AEGetParamDesc(message, keyDirectObject, typeAEList, &docList);

	if (err==noErr)
	{
		err = AECountItems( &docList, &itemsInList) ;
	}
	else
	{
		itemsInList = 0;
	}
	
	for (index = 1; index <= itemsInList; index++)
		if (err==noErr)
		{
			err = AEGetNthPtr( &docList, index, typeFSS, &keywd, &typeCode,
												 (Ptr)&theFSSpec, sizeof(theFSSpec), &actSize ) ;
			if( ( err == noErr ) && ( theSpecCount < kMaxSpecs ) )
			{
				theSpecList[ theSpecCount ] = theFSSpec;
				theSpecCount++;
			}
		}

	if (docList.dataHandle)
	{
		ignoreErr = AEDisposeDesc(&docList);
	}
	
	if( noErr == err )
	{
		WriteOutEachFrameOfSomeMovies( theSpecList, theSpecCount );
	}
	
	return err;
}

//
//	DoAEQuit
//
static pascal OSErr 
DoAEQuit(
	const AppleEvent * /*message*/,
	const AppleEvent * /*reply*/,
	long /*refcon*/ )			
{
	gDone = true;
	return noErr;
}


//
//	DoAEPrintDocuments
//
static pascal OSErr 
DoAEPrintDocuments(
	const AppleEvent * /*message*/, 
	AppleEvent * /*reply*/, 
	long /*refcon*/ )
{
	gDone = true;
	return errAEEventFailed;
} /* DoPrintDocuments */


//
//	InitAppleEvents
//
static void
InitAppleEvents(void)
{
	const long noRefCon = 0;
	OSErr aevtErr;
	
	aevtErr = AEInstallEventHandler( kCoreEventClass, kAEOpenApplication, NewAEEventHandlerProc(DoAEOpenApp), noRefCon, false) ;
	aevtErr = AEInstallEventHandler( kCoreEventClass, kAEOpenDocuments,   NewAEEventHandlerProc(DoAEOpenDocument), noRefCon, false) ;
	aevtErr = AEInstallEventHandler( kCoreEventClass, kAEPrintDocuments,  NewAEEventHandlerProc(DoAEPrintDocuments), noRefCon, false) ;
	aevtErr = AEInstallEventHandler( kCoreEventClass, kAEQuitApplication, NewAEEventHandlerProc(DoAEQuit), noRefCon, false) ;
}

//
//	WaitForAppleEvents
//
static void
WaitForAppleEvents()
{
	UInt32	patientTicks = TickCount() + 600;
	InitAppleEvents();
	SetCursor(&qd.arrow);
	
	gDone = false;
	while( !gDone )
	{
		EventRecord	event;
		Boolean		gotEvent;
		gotEvent = WaitNextEvent(everyEvent, &event, 60L, nil);
		if( gotEvent )
		{
			switch( event.what )
			{
				case kHighLevelEvent: 
					DoAppleEvent( &event );
					break;
			}
		}
		
		if( ( !gDone ) && ( TickCount() > patientTicks ) )
		{
				// Let's leave the loop if we haven't received any apple-events after 10 seconds.
				// NOTE: Make sure you have "Is High-Level Event Aware" turned on in the SIZE flags!
			PromptUserForAFileAndDoIt();
			gDone = true;
		}
	}
}


//====================================================================================
//
//	main
//
void main()
{
	OSErr		result;
	
	InitGraf(&qd.thePort);
	InitFonts();
	FlushEvents(0xffff,0);
	InitWindows();
	InitMenus();
	InitDialogs(0);
	TEInit();
	InitCursor();
	MaxApplZone();
	
	// The dialog manager doesn't like it if there are no menus around. 
	MacInsertMenu( NewMenu( 128,"\p" ), 0 );
	
	// You need some port set before you call NewMovie().
	{
		CGrafPort	port;
		OpenCPort( &port );
		SetPort( (GrafPtr)&port );
	}
	
	//---------------------------------------------------------------------------------------------
	//	Check to make sure at least QuickTime 3.0 is installed.  We depend on features
	//	that are only present in 3.0 or later.  
	//---------------------------------------------------------------------------------------------
	{
		long version;

		result = Gestalt(gestaltQuickTime,&version);
		if (result || version < 0x03000000)
			BailNil(0);
	}
	

	//---------------------------------------------------------------------------------------------
	//	Initialize the Movie Toolbox.
	//	Make sure you do this.  It's very easy to forget about.
	//---------------------------------------------------------------------------------------------
	
	result = EnterMovies();
	BailError(result);


	WaitForAppleEvents();

bail:;
}
